import sys
import importlib.resources
import json
from functools import lru_cache
from warnings import warn

from typing import IO, Any, Dict, Optional, Set

from emoji.unicode_codes.data_dict import STATUS, LANGUAGES

__all__ = [
    'get_emoji_by_name',
    'load_from_json',
    'EMOJI_DATA',
    'STATUS',
    'LANGUAGES',
]

_DEFAULT_KEYS = ('en', 'alias', 'E', 'status')  # The keys in emoji.json

_loaded_keys: Set[str] = set(
    _DEFAULT_KEYS
)  # Keep track of keys already loaded from json files to avoid loading them twice


@lru_cache(maxsize=4000)
def get_emoji_by_name(name: str, language: str) -> Optional[str]:
    """
    Find emoji by short-name in a specific language.
    Returns None if not found

    :param name: emoji short code e.g. ":banana:"
    :param language: language-code e.g. 'es', 'de', etc. or 'alias'
    """

    fully_qualified = STATUS['fully_qualified']

    if language == 'alias':
        for emj, data in EMOJI_DATA.items():
            if name in data.get('alias', []) and data['status'] <= fully_qualified:
                return emj
        language = 'en'

    for emj, data in EMOJI_DATA.items():
        if data.get(language) == name and data['status'] <= fully_qualified:
            return emj

    return None


class EmojiDataDict(Dict[str, Any]):
    """Replaces built-in-dict in the values of the EMOJI_DATA dict.
    Auto loads language data when accessing language data via
    key-access without prior loading of the language:
    e.g. EMOJI_DATA['👌']['fr'] will auto load French language and not throw
    a KeyError.
    Shows a deprecation warning explainging that `emoji.config.load_language()`
    should be used."""

    def __missing__(self, key: str) -> str:
        """Auto load language `key`, raises KeyError if language is no supported."""
        if key in LANGUAGES and key not in _loaded_keys:
            load_from_json(key)
            if key in self:
                warn(
                    f"""Use emoji.config.load_language('{key}') before accesing EMOJI_DATA[emj]['{key}'].
Accessing EMOJI_DATA[emj]['{key}'] without loading the language is deprecated.""",
                    DeprecationWarning,
                    stacklevel=3,
                )
                return self[key]  # type: ignore

        raise KeyError(key)


EMOJI_DATA: Dict[str, Dict[str, Any]]


def _open_file(name: str) -> IO[bytes]:
    if sys.version_info >= (3, 9):
        return importlib.resources.files('emoji.unicode_codes').joinpath(name).open('rb')
    else:
        return importlib.resources.open_binary('emoji.unicode_codes', name)


def _load_default_from_json():
    global EMOJI_DATA
    global _loaded_keys

    with _open_file('emoji.json') as f:
        EMOJI_DATA = dict(json.load(f, object_pairs_hook=EmojiDataDict))  # type: ignore
    _loaded_keys = set(_DEFAULT_KEYS)


def load_from_json(key: str):
    """Load values from the file 'emoji_{key}.json' into EMOJI_DATA"""

    if key in _loaded_keys:
        return

    if key not in LANGUAGES:
        raise NotImplementedError('Language not supported', key)

    with _open_file(f'emoji_{key}.json') as f:
        for emj, value in json.load(f).items():
            EMOJI_DATA[emj][key] = value  # type: ignore

    _loaded_keys.add(key)


_load_default_from_json()



"""Data containing all current emoji
Extracted from https://unicode.org/Public/emoji/latest/emoji-test.txt
and https://www.unicode.org/Public/UCD/latest/ucd/emoji/emoji-variation-sequences.txt
See utils/generate_emoji.py

+----------------+-------------+------------------+-------------------+
| Emoji Version  |    Date     | Unicode Version  | Data File Comment |
+----------------+-------------+------------------+-------------------+
| N/A            | 2010-10-11  | Unicode 6.0      | E0.6              |
| N/A            | 2014-06-16  | Unicode 7.0      | E0.7              |
| Emoji 1.0      | 2015-06-09  | Unicode 8.0      | E1.0              |
| Emoji 2.0      | 2015-11-12  | Unicode 8.0      | E2.0              |
| Emoji 3.0      | 2016-06-03  | Unicode 9.0      | E3.0              |
| Emoji 4.0      | 2016-11-22  | Unicode 9.0      | E4.0              |
| Emoji 5.0      | 2017-06-20  | Unicode 10.0     | E5.0              |
| Emoji 11.0     | 2018-05-21  | Unicode 11.0     | E11.0             |
| Emoji 12.0     | 2019-03-05  | Unicode 12.0     | E12.0             |
| Emoji 12.1     | 2019-10-21  | Unicode 12.1     | E12.1             |
| Emoji 13.0     | 2020-03-10  | Unicode 13.0     | E13.0             |
| Emoji 13.1     | 2020-09-15  | Unicode 13.0     | E13.1             |
| Emoji 14.0     | 2021-09-14  | Unicode 14.0     | E14.0             |
| Emoji 15.0     | 2022-09-13  | Unicode 15.0     | E15.0             |
| Emoji 15.1     | 2023-09-12  | Unicode 15.1     | E15.1             |
| Emoji 16.0     | 2024-09-10  | Unicode 16.0     | E16.0             |

               http://www.unicode.org/reports/tr51/#Versioning

"""

__all__ = ['STATUS', 'LANGUAGES']

from typing import Any, Dict, List


component = 1
fully_qualified = 2
minimally_qualified = 3
unqualified = 4

STATUS: Dict[str, int] = {
    'component': component,
    'fully_qualified': fully_qualified,
    'minimally_qualified': minimally_qualified,
    'unqualified': unqualified,
}

LANGUAGES: List[str] = [
    'en',
    'es',
    'ja',
    'ko',
    'pt',
    'it',
    'fr',
    'de',
    'fa',
    'id',
    'zh',
    'ru',
    'tr',
    'ar',
]


# The following is only an example of how the EMOJI_DATA dict is structured.
# The real data is loaded from the json files at runtime, see unicode_codes/__init__.py
EMOJI_DATA: Dict[str, Dict[str, Any]] = {
    '\U0001f947': {  # 🥇
        'en': ':1st_place_medal:',
        'status': fully_qualified,
        'E': 3,
        'de': ':goldmedaille:',
        'es': ':medalla_de_oro:',
        'fr': ':médaille_d’or:',
        'ja': ':金メダル:',
        'ko': ':금메달:',
        'pt': ':medalha_de_ouro:',
        'it': ':medaglia_d’oro:',
        'fa': ':مدال_طلا:',
        'id': ':medali_emas:',
        'zh': ':金牌:',
        'ru': ':золотая_медаль:',
        'tr': ':birincilik_madalyası:',
        'ar': ':ميدالية_مركز_أول:',
    },
    '\U0001f948': {  # 🥈
        'en': ':2nd_place_medal:',
        'status': fully_qualified,
        'E': 3,
        'de': ':silbermedaille:',
        'es': ':medalla_de_plata:',
        'fr': ':médaille_d’argent:',
        'ja': ':銀メダル:',
        'ko': ':은메달:',
        'pt': ':medalha_de_prata:',
        'it': ':medaglia_d’argento:',
        'fa': ':مدال_نقره:',
        'id': ':medali_perak:',
        'zh': ':银牌:',
        'ru': ':серебряная_медаль:',
        'tr': ':ikincilik_madalyası:',
        'ar': ':ميدالية_مركز_ثان:',
    },
    '\U0001f949': {  # 🥉
        'en': ':3rd_place_medal:',
        'status': fully_qualified,
        'E': 3,
        'de': ':bronzemedaille:',
        'es': ':medalla_de_bronce:',
        'fr': ':médaille_de_bronze:',
        'ja': ':銅メダル:',
        'ko': ':동메달:',
        'pt': ':medalha_de_bronze:',
        'it': ':medaglia_di_bronzo:',
        'fa': ':مدال_برنز:',
        'id': ':medali_perunggu:',
        'zh': ':铜牌:',
        'ru': ':бронзовая_медаль:',
        'tr': ':üçüncülük_madalyası:',
        'ar': ':ميدالية_مركز_ثالث:',
    },
    '\U0001f18e': {  # 🆎
        'en': ':AB_button_(blood_type):',
        'status': fully_qualified,
        'E': 0.6,
        'alias': [':ab:', ':ab_button_blood_type:'],
        'de': ':großbuchstaben_ab_in_rotem_quadrat:',
        'es': ':grupo_sanguíneo_ab:',
        'fr': ':groupe_sanguin_ab:',
        'ja': ':血液型ab型:',
        'ko': ':에이비형:',
        'pt': ':botão_ab_(tipo_sanguíneo):',
        'it': ':gruppo_sanguigno_ab:',
        'fa': ':دکمه_آ_ب_(گروه_خونی):',
        'id': ':tombol_ab_(golongan_darah):',
        'zh': ':AB型血:',
        'ru': ':IV_группа_крови:',
        'tr': ':ab_düğmesi_(kan_grubu):',
        'ar': ':زر_ab_(فئة_الدم):',
    },
    '\U0001f3e7': {  # 🏧
        'en': ':ATM_sign:',
        'status': fully_qualified,
        'E': 0.6,
        'alias': [':atm:', ':atm_sign:'],
        'de': ':symbol_geldautomat:',
        'es': ':señal_de_cajero_automático:',
        'fr': ':distributeur_de_billets:',
        'ja': ':atm:',
        'ko': ':에이티엠:',
        'pt': ':símbolo_de_caixa_automático:',
        'it': ':simbolo_dello_sportello_bancomat:',
        'fa': ':نشان_عابربانک:',
        'id': ':tanda_atm:',
        'zh': ':取款机:',
        'ru': ':значок_банкомата:',
        'tr': ':atm_işareti:',
        'ar': ':علامة_ماكينة_صرف_آلي:',
    },
    '\U0001f170\U0000fe0f': {  # 🅰️
        'en': ':A_button_(blood_type):',
        'status': fully_qualified,
        'E': 0.6,
        'alias': [':a:', ':a_button_blood_type:'],
        'variant': True,
        'de': ':großbuchstabe_a_in_rotem_quadrat:',
        'es': ':grupo_sanguíneo_a:',
        'fr': ':groupe_sanguin_a:',
        'ja': ':血液型a型:',
        'ko': ':에이형:',
        'pt': ':botão_a_(tipo_sanguíneo):',
        'it': ':gruppo_sanguigno_a:',
        'fa': ':دکمه_آ_(گروه_خونی):',
        'id': ':tombol_a_(golongan_darah):',
        'zh': ':A型血:',
        'ru': ':ii_группа_крови:',
        'tr': ':a_düğmesi_(kan_grubu):',
        'ar': ':زر_a:',
    },
    '\U0001f170': {  # 🅰
        'en': ':A_button_(blood_type):',
        'status': unqualified,
        'E': 0.6,
        'alias': [':a:', ':a_button_blood_type:'],
        'variant': True,
        'de': ':großbuchstabe_a_in_rotem_quadrat:',
        'es': ':grupo_sanguíneo_a:',
        'fr': ':groupe_sanguin_a:',
        'ja': ':血液型a型:',
        'ko': ':에이형:',
        'pt': ':botão_a_(tipo_sanguíneo):',
        'it': ':gruppo_sanguigno_a:',
        'fa': ':دکمه_آ_(گروه_خونی):',
        'id': ':tombol_a_(golongan_darah):',
        'zh': ':A型血:',
        'ru': ':II_группа_крови:',
        'tr': ':a_düğmesi_(kan_grubu):',
        'ar': ':زر_a:',
    },
    '\U0001f1e6\U0001f1eb': {  # 🇦🇫
        'en': ':Afghanistan:',
        'status': fully_qualified,
        'E': 2,
        'alias': [':flag_for_Afghanistan:', ':afghanistan:'],
        'de': ':flagge_afghanistan:',
        'es': ':bandera_afganistán:',
        'fr': ':drapeau_afghanistan:',
        'ja': ':旗_アフガニスタン:',
        'ko': ':깃발_아프가니스탄:',
        'pt': ':bandeira_afeganistão:',
        'it': ':bandiera_afghanistan:',
        'fa': ':پرچم_افغانستان:',
        'id': ':bendera_afganistan:',
        'zh': ':阿富汗:',
        'ru': ':флаг_Афганистан:',
        'tr': ':bayrak_afganistan:',
        'ar': ':علم_أفغانستان:',
    },
    '\U0001f1e6\U0001f1f1': {  # 🇦🇱
        'en': ':Albania:',
        'status': fully_qualified,
        'E': 2,
        'alias': [':flag_for_Albania:', ':albania:'],
        'de': ':flagge_albanien:',
        'es': ':bandera_albania:',
        'fr': ':drapeau_albanie:',
        'ja': ':旗_アルバニア:',
        'ko': ':깃발_알바니아:',
        'pt': ':bandeira_albânia:',
        'it': ':bandiera_albania:',
        'fa': ':پرچم_آلبانی:',
        'id': ':bendera_albania:',
        'zh': ':阿尔巴尼亚:',
        'ru': ':флаг_Албания:',
        'tr': ':bayrak_arnavutluk:',
        'ar': ':علم_ألبانيا:',
    },
    '\U0001f1e9\U0001f1ff': {  # 🇩🇿
        'en': ':Algeria:',
        'status': fully_qualified,
        'E': 2,
        'alias': [':flag_for_Algeria:', ':algeria:'],
        'de': ':flagge_algerien:',
        'es': ':bandera_argelia:',
        'fr': ':drapeau_algérie:',
        'ja': ':旗_アルジェリア:',
        'ko': ':깃발_알제리:',
        'pt': ':bandeira_argélia:',
        'it': ':bandiera_algeria:',
        'fa': ':پرچم_الجزایر:',
        'id': ':bendera_aljazair:',
        'zh': ':阿尔及利亚:',
        'ru': ':флаг_Алжир:',
        'tr': ':bayrak_cezayir:',
        'ar': ':علم_الجزائر:',
    },
    '\U0001f1e6\U0001f1f8': {  # 🇦🇸
        'en': ':American_Samoa:',
        'status': fully_qualified,
        'E': 2,
        'alias': [':flag_for_American_Samoa:', ':american_samoa:'],
        'de': ':flagge_amerikanisch-samoa:',
        'es': ':bandera_samoa_americana:',
        'fr': ':drapeau_samoa_américaines:',
        'ja': ':旗_米領サモア:',
        'ko': ':깃발_아메리칸_사모아:',
        'pt': ':bandeira_samoa_americana:',
        'it': ':bandiera_samoa_americane:',
        'fa': ':پرچم_ساموآی_امریکا:',
        'id': ':bendera_samoa_amerika:',
        'zh': ':美属萨摩亚:',
        'ru': ':флаг_Американское_Самоа:',
        'tr': ':bayrak_amerikan_samoası:',
        'ar': ':علم_ساموا_الأمريكية:',
    },
}



"""Tests for emoji that consist of multiple emoji joined with a u200D (ZWJ - zero width joiner)
This file contains tests when the ZWJ is kept in place by demojize/replace_emoji.
See test_zwj_remove.py for tests when the ZWJ is removed.
"""

from typing import Any, Dict
import emoji


def ascii(s: str) -> str:
    # return escaped Code points \U000AB123
    return s.encode('unicode-escape').decode()


def test_non_rgi_zwj_demojize():
    # These emoji are non-RGI ZWJ sequences. They should be decoded by demojize to their constituents.
    emoji.config.demojize_keep_zwj = True

    assert (
        emoji.demojize(
            '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
        )
        == ':man:\u200d:woman_dark_skin_tone:\u200d:girl_light_skin_tone:\u200d:boy_medium-dark_skin_tone:'
    )

    assert (
        emoji.demojize(
            '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
        )
        == ':man_dark_skin_tone:\u200d:woman_dark_skin_tone:\u200d:girl_light_skin_tone:\u200d:boy_medium-dark_skin_tone:'
    )

    assert (
        emoji.demojize(
            '\U0001f468\U0001f3ff\u200d\U0001f469\u200d\U0001f467\U0001f3fb\u200d\U0001f466'
        )
        == ':man_dark_skin_tone:\u200d:family_woman_girl::light_skin_tone:\u200d:boy:'
    )

    # https://www.unicode.org/Public/15.0.0/ucd/auxiliary/GraphemeBreakTest-15.0.0d1.html#s19
    assert (
        emoji.demojize('\U0001f3ff\U0001f476\u200d\U0001f6d1')
        == ':dark_skin_tone::baby:\u200d:stop_sign:'
    )

    # https://www.unicode.org/Public/15.0.0/ucd/auxiliary/GraphemeBreakTest-15.0.0d1.html#s20
    # Check that \u200d is kept, if it not part of a valid ZWJ emoji
    assert (
        emoji.demojize('\U0001f476\U0001f3ff\u0308\u200d\U0001f476\U0001f3ff')
        == ':baby_dark_skin_tone:\u0308\u200d:baby_dark_skin_tone:'
    )


def test_malformed_zwj_demojize():
    # These sequences are malformed in the sense that they are neither RGI nor non-RGI ZWJ sequences
    # Check that still all the emoji are decoded
    emoji.config.demojize_keep_zwj = True

    result = emoji.demojize(
        '\U0001f468\u200d\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
    )
    assert '\U0001f468' not in result
    assert '\U0001f3ff' not in result
    assert '\U0001f467\U0001f3fb' not in result
    assert '\U0001f466\U0001f3fe' not in result

    result = emoji.demojize(
        '\U0001f3ff\u200d\U0001f468\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
    )
    assert '\U0001f468' not in result
    assert '\U0001f3ff' not in result
    assert '\U0001f467\U0001f3fb' not in result
    assert '\U0001f466\U0001f3fe' not in result

    # Mix of emoji and other characters
    # Check that still all the emoji are decoded
    result = emoji.demojize(
        '\U0001f468\U0001f3ff\u200dabcd\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
    )
    assert '\U0001f468\U0001f3ff' not in result
    assert '\U0001f467\U0001f3fb' not in result
    assert '\U0001f466\U0001f3fe' not in result
    assert result.count('\u200d') == 3
    assert ':man_dark_skin_tone:' in result
    assert 'abcd' in result
    assert ':girl_light_skin_tone:' in result
    assert ':boy_medium-dark_skin_tone:' in result

    result = emoji.demojize(
        'W\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
    )
    assert '\U0001f469\U0001f3ff' not in result
    assert '\U0001f467\U0001f3fb' not in result
    assert '\U0001f466\U0001f3fe' not in result
    assert result.count('\u200d') == 3
    assert result.startswith('W')
    assert ':woman_dark_skin_tone:' in result
    assert ':girl_light_skin_tone:' in result
    assert ':boy_medium-dark_skin_tone:' in result

    result = emoji.demojize('\U0001f9d1\U0001f3fe\u200d\u200d\U0001f393')
    assert '\U0001f9d1\U0001f3fe' not in result
    assert '\U0001f393' not in result
    assert result.count('\u200d') == 2
    assert ':person_medium-dark_skin_tone:' in result
    assert ':graduation_cap:' in result


def test_non_rgi_zwj_replace():
    emoji.config.replace_emoji_keep_zwj = True

    assert (
        emoji.replace_emoji(
            '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
            '',
        )
        == ''
    )
    assert (
        emoji.replace_emoji(
            '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
            'X',
        )
        == 'X'
    )

    assert (
        emoji.replace_emoji(
            '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
            '',
        )
        == ''
    )
    assert (
        emoji.replace_emoji(
            '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
            'X',
        )
        == 'X'
    )

    assert (
        emoji.replace_emoji(
            '\U0001f468\U0001f3ff\u200d\U0001f469\u200d\U0001f467\U0001f3fb\u200d\U0001f466',
            '',
        )
        == ''
    )
    assert (
        emoji.replace_emoji(
            '\U0001f468\U0001f3ff\u200d\U0001f469\u200d\U0001f467\U0001f3fb\u200d\U0001f466',
            'X',
        )
        == 'XX'
    )

    assert emoji.replace_emoji('\U0001f3ff\U0001f476\u200d\U0001f6d1', 'X') == 'XX'

    assert (
        emoji.replace_emoji('\U0001f476\U0001f3ff\u0308\u200d\U0001f476\U0001f3ff', 'X')
        == 'X\u0308\u200dX'
    )

    # Replace with different length
    index = [0]

    def replace_f(e: str, emoji_data: Dict[str, Any]) -> str:
        index[0] += 1
        if index[0] % 2 == 0:
            return 'X'
        else:
            return 'yyy'

    assert (
        emoji.replace_emoji(
            'A\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe\U0001f468\U0001f3fb\u200d\U0001f469\U0001f3fc\u200d\U0001f467\U0001f3fe\U0001f469\U0001f3feB',
            replace_f,
        )
        == 'AyyyXyyyB'
    )


def test_non_rgi_zwj_reversible():
    emoji.config.demojize_keep_zwj = True
    for emj in [
        '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
        '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
        '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3fe\u200d\U0001f466\U0001f3fd\u200d\U0001f467\U0001f3fb',
        'Test \U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe abc',
        '\U0001f468\U0000200d\U00002708\U0000fe0f\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
        '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3fe\u200d\U0001f466\U0001f3fd\u200d\U0001f467\U0001f3fb\U0001f64e\U0001f3fd\U0000200d\U00002642\U0000fe0f',
        '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe',
        '\U0001f3ff\U0001f476\u200d\U0001f6d1',
    ]:
        assert emoji.emojize(emoji.demojize(emj)) == emj



"""Unittests for emoji.analyze()"""

import emoji


def test_analyze():
    assert list(emoji.analyze('')) == []

    assert list(emoji.analyze('', non_emoji=True)) == []

    assert list(emoji.analyze('abc')) == []

    assert list(emoji.analyze('abc', non_emoji=True)) == [
        ('a', 'a'),
        ('b', 'b'),
        ('c', 'c'),
    ]

    result = list(emoji.analyze('abc\U0001f472'))
    assert len(result) == 1
    assert not isinstance(result[0].value, str)
    assert result[0].value.emoji == '\U0001f472'

    result = list(emoji.analyze('abc\U0001f472', non_emoji=True))
    assert result[0].value == 'a'
    assert not isinstance(result[3].value, str)
    assert result[3].value.emoji == '\U0001f472'

    result = list(emoji.analyze('\U0001f477\U0001f3fb\U0000200d\U00002640'))
    assert len(result) == 1
    assert not isinstance(result[0].value, str)
    assert result[0].value.emoji == '\U0001f477\U0001f3fb\U0000200d\U00002640'

    result = list(
        emoji.analyze('\U0001f477\U0001f3fc\U0001f477\U0001f3fb\U0000200d\U00002640')
    )
    assert len(result) == 2
    assert not isinstance(result[0].value, str)
    assert result[0].value.emoji == '\U0001f477\U0001f3fc'
    assert not isinstance(result[1].value, str)
    assert result[1].value.emoji == '\U0001f477\U0001f3fb\U0000200d\U00002640'


def test_analyze_non_rgi_zwj():
    result = list(
        emoji.analyze(
            '\U0001f468\U0001f3ff\U0000200d\U0001f469\U0001f3fb\U0000200d\U0001f467\U0001f3fd'
        )
    )
    assert len(result) == 1
    assert not isinstance(result[0].value, str)
    assert (
        result[0].value.emoji
        == '\U0001f468\U0001f3ff\U0000200d\U0001f469\U0001f3fb\U0000200d\U0001f467\U0001f3fd'
    )

    result = list(
        emoji.analyze(
            '\U0001f468\U0001f3ff\U0000200d\U0001f469\U0001f3fb\U0000200d\U0001f467\U0001f3fd',
            join_emoji=False,
        )
    )
    assert len(result) == 3
    assert not isinstance(result[0].value, str)
    assert result[0].value.emoji == '\U0001f468\U0001f3ff'
    assert not isinstance(result[1].value, str)
    assert result[1].value.emoji == '\U0001f469\U0001f3fb'
    assert not isinstance(result[2].value, str)
    assert result[2].value.emoji == '\U0001f467\U0001f3fd'

    result = list(
        emoji.analyze(
            '\U0001f468\U0001f3ff\U0000200d\U0001f469\U0001f3fb\U0000200d\U0001f467\U0001f3fdx',
            join_emoji=False,
            non_emoji=True,
        )
    )
    assert len(result) == 6
    assert not isinstance(result[0].value, str)
    assert result[0].value.emoji == '\U0001f468\U0001f3ff'
    assert result[1].value == '\U0000200d'
    assert not isinstance(result[2].value, str)
    assert result[2].value.emoji == '\U0001f469\U0001f3fb'
    assert result[3].value == '\U0000200d'
    assert not isinstance(result[4].value, str)
    assert result[4].value.emoji == '\U0001f467\U0001f3fd'
    assert result[5].value == 'x'

    result = list(
        emoji.analyze(
            '\U0001f468\U0001f3ff\U0000200d\U0001f469\U0001f3fb\U0000200d\U0001f467\U0001f3fdx',
            join_emoji=True,
            non_emoji=True,
        )
    )
    assert len(result) == 2
    assert not isinstance(result[0].value, str)
    assert (
        result[0].value.emoji
        == '\U0001f468\U0001f3ff\U0000200d\U0001f469\U0001f3fb\U0000200d\U0001f467\U0001f3fd'
    )
    assert result[1].value == 'x'

    result = list(emoji.analyze('\u200d🦷\u200d🦷'))
    assert len(result) == 1
    assert isinstance(result[0].value, emoji.EmojiMatchZWJNonRGI)

    result = list(emoji.analyze('\u200d🦷\u200d🦷', join_emoji=False))
    assert len(result) == 2
    assert all(isinstance(token.value, emoji.EmojiMatch) for token in result)

    result = list(emoji.analyze('\u200d🦷\u200d🦷', join_emoji=False, non_emoji=True))
    assert len(result) == 4
    assert result[0].value == '\u200d'
    assert isinstance(result[1].value, emoji.EmojiMatch)
    assert result[2].value == '\u200d'
    assert isinstance(result[3].value, emoji.EmojiMatch)


def test_emoji_match():
    s = 'a\U0001f309b'
    token = next(emoji.analyze(s))
    assert isinstance(token, emoji.Token)

    assert token.chars == s[1:-1]

    match = token.value

    assert isinstance(match, emoji.EmojiMatch)
    assert match.emoji == s[1:-1]
    assert match.start == 1
    assert match.end == 2
    assert not match.is_zwj()
    assert match.split() == match
    assert str(match).startswith('EmojiMatch(')

    s = 'a\U0001f468\U0000200d\U0001f467b'
    token = next(emoji.analyze(s))
    assert isinstance(token, emoji.Token)

    assert token.chars == s[1:-1]

    match = token.value

    assert isinstance(match, emoji.EmojiMatch)
    assert match.emoji == s[1:-1]
    assert match.start == 1
    assert match.end == 4
    assert match.is_zwj()
    assert str(match).startswith('EmojiMatch(')

    splitted = match.split()
    assert isinstance(splitted, emoji.EmojiMatchZWJ)
    assert splitted.emoji == s[1:-1]
    assert splitted.start == 1
    assert splitted.end == 4
    assert splitted.is_zwj()
    assert splitted.join() == s[1:-1]
    assert splitted.split() == splitted
    assert str(splitted).startswith('EmojiMatchZWJ(')
    assert len(splitted.emojis) == 2

    man, woman = splitted.emojis

    assert isinstance(man, emoji.EmojiMatch)
    assert man.start == 1
    assert man.end == 2

    assert isinstance(woman, emoji.EmojiMatch)
    assert woman.start == 3
    assert woman.end == 4



"""Unittests for versions in EMOJI_DATA"""

from typing import Any, Dict, List
import pytest
import emoji
from testutils import all_language_packs


def test_emoji_versions_complete_emojize():
    # Check that every emoji has a valid version
    replacement = '<3'
    for lang_code, emoji_pack in all_language_packs():
        for name in emoji_pack.keys():
            version: List[float] = []

            def f(e: str, d: Dict[str, Any]) -> str:
                v = d['E']
                n = d[lang_code]
                assert n == name
                assert isinstance(v, (int, float))
                assert v >= 0.6
                version.append(v)
                return replacement

            r = emoji.emojize(name, language=lang_code, version=0.0, handle_version=f)
            assert len(version) == 1
            assert r == replacement


def test_emoji_versions_complete_demojize():
    # Check that every emoji has a valid version
    for lang_code, emoji_pack in all_language_packs():
        for name in emoji_pack.keys():
            version: List[float] = []

            def f(e: str, d: Dict[str, Any]) -> str:
                v = d['E']
                assert isinstance(v, (int, float))
                assert v >= 0.6
                version.append(v)
                return ''

            emoji.demojize(
                emoji.emojize(name, language=lang_code),
                language=lang_code,
                version=0.0,
                handle_version=f,
            )
            if len(version) != 1:
                print(name)
                print(
                    emoji.emojize(name, language=lang_code)
                    .encode('unicode-escape')
                    .decode('utf-8')
                )
                print(
                    emoji.demojize(
                        emoji.emojize(name, language=lang_code), language=lang_code
                    )
                    .encode('unicode-escape')
                    .decode('utf-8')
                )
            assert len(version) == 1


def test_method_version():
    # Test method "emoji.version()"

    assert emoji.version(':snake:') == 0.6
    assert emoji.version('\U0001f40d') == 0.6

    assert emoji.version(':brain:') == 5
    assert emoji.version('\U0001f9e0') == 5
    assert emoji.version(':cerebro:') == 5

    assert emoji.version('prefix :people_hugging: suffix') == 13
    assert emoji.version('prefix \U0001fac2 suffix') == 13

    assert emoji.version('\U0001f611') == 1
    assert emoji.version(':expressionless_face:') == 1
    assert emoji.version(':expressionless:') == 1

    assert emoji.version("':pouring_liquid::people_hugging:") == 14
    assert emoji.version('\U0001fad7\U0001fac2') == 14

    assert emoji.emojize(':123:', version=5, variant='text_type') == ':123:'

    with pytest.raises(ValueError):
        emoji.version('test')

    with pytest.raises(ValueError):
        emoji.version("':snak:")


def test_method_replace_version():
    # Test method "emoji.replace_emoji(string, version=X.Y)"

    assert (
        emoji.replace_emoji(
            '\U0001f40d\U0001f9e0\U0001fac2\U0001fad7\U0001fac2', version=6
        )
        == '\U0001f40d\U0001f9e0'
    )

    assert emoji.replace_emoji('Hi, I am fine. 😁', version=0) == 'Hi, I am fine. '
    assert emoji.replace_emoji('Hi', version=0) == 'Hi'
    assert emoji.replace_emoji('Hello 🇫🇷👌', version=0) == 'Hello '
    assert (
        emoji.replace_emoji(
            'Hello 🇫🇷👌',
            'x',
            version=0,
        )
        == 'Hello xx'
    )
    assert (
        emoji.replace_emoji(
            'Hello 🇫🇷👌',
            'x',
            version=1,
        )
        == 'Hello 🇫🇷👌'
    )

    def replace(emj: str, data: Dict[str, Any]) -> str:
        assert emj in ['🇫🇷', '👌']
        return 'x'

    assert emoji.replace_emoji('Hello 🇫🇷👌', replace, version=0.1) == 'Hello xx'

    assert (
        emoji.replace_emoji('A 🦖 is eating a 🥐', replace='', version=5.0)
        == 'A 🦖 is eating a 🥐'
    )



"""Tests for emoji that consist of multiple emoji joined with a u200D (ZWJ - zero width joiner)
This file contains tests that are irrespective of keeping/removing the ZWJ.
See test_zwj_remove.py for tests when the ZWJ is removed.
See test_zwj_keep.py for tests when the ZWJ is kept.
"""

import emoji


def test_non_rgi_zwj_emoji_list():
    matches = emoji.emoji_list(
        '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
    )

    assert matches[0]['match_start'] == 0
    assert matches[0]['match_end'] == 1
    assert matches[0]['emoji'] == '\U0001f468'

    assert matches[1]['match_start'] == 2
    assert matches[1]['match_end'] == 4
    assert matches[1]['emoji'] == '\U0001f469\U0001f3ff'

    assert matches[2]['match_start'] == 5
    assert matches[2]['match_end'] == 7
    assert matches[2]['emoji'] == '\U0001f467\U0001f3fb'

    assert matches[3]['match_start'] == 8
    assert matches[3]['match_end'] == 10
    assert matches[3]['emoji'] == '\U0001f466\U0001f3fe'

    matches = emoji.emoji_list(
        'abc\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3feyxz'
    )

    assert matches[0]['match_start'] == 3
    assert matches[0]['match_end'] == 4
    assert matches[0]['emoji'] == '\U0001f468'

    assert matches[1]['match_start'] == 5
    assert matches[1]['match_end'] == 7
    assert matches[1]['emoji'] == '\U0001f469\U0001f3ff'

    assert matches[2]['match_start'] == 8
    assert matches[2]['match_end'] == 10
    assert matches[2]['emoji'] == '\U0001f467\U0001f3fb'

    assert matches[3]['match_start'] == 11
    assert matches[3]['match_end'] == 13
    assert matches[3]['emoji'] == '\U0001f466\U0001f3fe'

    matches = emoji.emoji_list(
        '\U0001f9d1\U0000200d\U0001f9b3abcyxz\U0001fac5\U0001f468\U0001f3fd\u200d\U0001f467\U0001f3fc\u200d\U0001f467\U0001f3fe'
    )

    assert matches[0]['match_start'] == 0
    assert matches[0]['match_end'] == 3
    assert matches[0]['emoji'] == '\U0001f9d1\U0000200d\U0001f9b3'

    assert matches[1]['match_start'] == 9
    assert matches[1]['match_end'] == 10
    assert matches[1]['emoji'] == '\U0001fac5'

    assert matches[2]['match_start'] == 10
    assert matches[2]['match_end'] == 12
    assert matches[2]['emoji'] == '\U0001f468\U0001f3fd'

    assert matches[3]['match_start'] == 13
    assert matches[3]['match_end'] == 15
    assert matches[3]['emoji'] == '\U0001f467\U0001f3fc'

    assert matches[4]['match_start'] == 16
    assert matches[4]['match_end'] == 18
    assert matches[4]['emoji'] == '\U0001f467\U0001f3fe'


def test_non_rgi_zwj_demojize():
    result = emoji.demojize(
        '\U0001f9d1\U0001f3fc\U0000200d\U0001f3a8\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
    )
    assert '\U0001f9d1\U0001f3fc\U0000200d\U0001f3a8' not in result
    assert (
        '\U0001f468\u200d\U0001f469\U0001f3ff\u200d\U0001f467\U0001f3fb\u200d\U0001f466\U0001f3fe'
        not in result
    )
    assert ':artist_medium-light_skin_tone:' in result

    result = emoji.demojize(
        'Test \U0001f6b5\U0001f3ff\U0000200d\U00002642\U0000fe0f abc \U0001f468\U0001f3ff\u200d\U0001f469\U0001f3fe\u200d\U0001f466\U0001f3fd\u200d\U0001f467\U0001f3fb'
    )
    assert '\U0001f6b5\U0001f3ff\U0000200d\U00002642\U0000fe0f' not in result
    assert (
        '\U0001f468\U0001f3ff\u200d\U0001f469\U0001f3fe\u200d\U0001f466\U0001f3fd\u200d\U0001f467\U0001f3fb'
        not in result
    )
    assert ':man_mountain_biking_dark_skin_tone:' in result


def test_malformed_zwj_no_emoji():
    s = '\u200d'
    assert emoji.replace_emoji(s) == s

    s = '\u200d\u200d'
    assert emoji.replace_emoji(s) == s

    s = '\u200d\u200d\u200d'
    assert emoji.replace_emoji(s) == s

    s = 'Has\u200din the middle'
    assert emoji.replace_emoji(s) == s

    s = '\u200dStarts With'
    assert emoji.replace_emoji(s) == s

    s = 'Ends With\u200d'
    assert emoji.replace_emoji(s) == s

    s = 'Multiple\u200d\u200d\u200din the middle'
    assert emoji.replace_emoji(s) == s

    s = '\u200d\u200dStarts With two'
    assert emoji.replace_emoji(s) == s

    s = '\u200d\u200d\u200dStarts With three'
    assert emoji.replace_emoji(s) == s

    s = 'Ends With two\u200d\u200d'
    assert emoji.replace_emoji(s) == s

    s = 'Ends With three\u200d\u200d\u200d'
    assert emoji.replace_emoji(s) == s


def test_malformed_zwj_mixed_with_emoji():
    emoji.config.demojize_keep_zwj = True  # Restore default config value

    i = 'Has🦷\u200din the middle'
    o = 'Has:tooth:\u200din the middle'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Has\u200d🦷in the middle'
    o = 'Has\u200d:tooth:in the middle'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d🦷Starts with'
    o = '\u200d:tooth:Starts with'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '🦷\u200dStarts with'
    o = ':tooth:\u200dStarts with'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Ends with \u200d🦷'
    o = 'Ends with \u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Ends with 🦷\u200d'
    o = 'Ends with :tooth:\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Multiple 🦷\u200d🦷\u200d in the middle'
    o = 'Multiple :tooth:\u200d:tooth:\u200d in the middle'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Multiple 🦷🦷\u200d\u200d in the middle'
    o = 'Multiple :tooth::tooth:\u200d\u200d in the middle'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Multiple \u200d\u200d🦷🦷 in the middle'
    o = 'Multiple \u200d\u200d:tooth::tooth: in the middle'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d🦷Starts with two'
    o = '\u200d\u200d:tooth:Starts with two'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d\u200d🦷Starts with three'
    o = '\u200d\u200d\u200d:tooth:Starts with three'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Ends with two \u200d\u200d🦷'
    o = 'Ends with two \u200d\u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Ends with two 🦷\u200d\u200d'
    o = 'Ends with two :tooth:\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Ends with three \u200d\u200d\u200d🦷'
    o = 'Ends with three \u200d\u200d\u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = 'Ends with three 🦷\u200d\u200d\u200d'
    o = 'Ends with three :tooth:\u200d\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '🦷\u200d'
    o = ':tooth:\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d🦷'
    o = '\u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d🦷'
    o = '\u200d\u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '🦷\u200d\u200d'
    o = ':tooth:\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d🦷\u200d'
    o = '\u200d:tooth:\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d🦷\u200d\u200d'
    o = '\u200d\u200d:tooth:\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d\u200d🦷\u200d\u200d'
    o = '\u200d\u200d\u200d:tooth:\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d🦷\u200d\u200d\u200d'
    o = '\u200d\u200d:tooth:\u200d\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '🦷\u200d\u200d🦷\u200d\u200d\u200d'
    o = ':tooth:\u200d\u200d:tooth:\u200d\u200d\u200d'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d🦷🦷\u200d\u200d\u200d🦷'
    o = '\u200d\u200d:tooth::tooth:\u200d\u200d\u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'

    i = '\u200d\u200d🦷\u200d\u200d\u200d🦷'
    o = '\u200d\u200d:tooth:\u200d\u200d\u200d:tooth:'
    assert emoji.demojize(i) == o, f'{i!r} != {o!r}'



"""Unittests for the big dict of dicts containing all emoji"""

from typing import Set, Dict
import re
import emoji

from testutils import load_all_languages as load_all_languages


def test_all_languages_list(load_all_languages):  # type:ignore
    """Compare all language keys in EMOJI_DATA with the emoji.LANGUAGES list"""

    langs: Set[str] = set()
    for item in emoji.EMOJI_DATA.values():
        langs.update(item.keys())
    all_languages = {lang for lang in langs if len(lang) == 2 and lang.lower() == lang}

    assert set(emoji.LANGUAGES) == all_languages


def test_emoji_versions():
    """Check that every emoji has a valid version"""
    for item in emoji.EMOJI_DATA.values():
        assert 'E' in item
        v = item['E']
        assert isinstance(v, (int, float))
        assert v >= 0.6


def check_duplicate_names(lang: str):
    """Check that there are no duplicate names in the fully_qualified except for different variants"""
    seen: Dict[str, int] = {}
    for item in emoji.EMOJI_DATA.values():
        if item['status'] > emoji.STATUS['fully_qualified']:
            continue

        if lang not in item:
            continue

        name = item[lang]
        if name in seen and 'variant' in item:
            seen[name] += 1
        else:
            assert name not in seen
            seen[name] = 0


def test_duplicate_names(load_all_languages):  # type:ignore
    """Check that there are no duplicate names in the fully_qualified except for different variants"""
    for lang in emoji.LANGUAGES:
        check_duplicate_names(lang)


def test_name_valid(load_all_languages):  # type:ignore
    """Check that every name starts with colons and does not contain other colons or whitespace"""

    pattern = re.compile(r':[^:\s]+:')
    for item in emoji.EMOJI_DATA.values():
        for lang in emoji.LANGUAGES:
            if lang in item:
                name = item[lang]
                assert pattern.match(name)



"""Unittests for emoji.core"""

import random
import re

from typing import Any, Callable, Dict, List, Literal, Tuple, Union
import pytest

import emoji

from testutils import (
    ascii,
    normalize,
    all_language_packs,
    all_language_and_alias_packs,
    get_emoji_unicode_dict,
    load_all_languages as load_all_languages,
)

try:
    from typeguard import suppress_type_checks  # type:ignore
except ImportError:
    from contextlib import nullcontext as suppress_type_checks  # type:ignore


def test_emojize_name_only(load_all_languages):  # type:ignore
    # Check that the regular expression emoji.core._EMOJI_NAME_PATTERN contains all the necesseary characters
    from emoji.core import _EMOJI_NAME_PATTERN  # pyright: ignore [reportPrivateUsage]

    pattern = re.compile('[^%s]' % (_EMOJI_NAME_PATTERN,))

    for lang_code, emoji_pack in all_language_and_alias_packs():
        for name_in_db in emoji_pack.keys():
            pairs = [
                ('Form from EMOJI_DATA', name_in_db),
                ('NFKC', normalize('NFKC', name_in_db)),
                ('NFKD', normalize('NFKD', name_in_db)),
                ('NFD', normalize('NFD', name_in_db)),
                ('NFC', normalize('NFC', name_in_db)),
            ]
            for form, name in pairs:
                actual = emoji.emojize(name, language=lang_code)
                expected = emoji_pack[name_in_db]

                if expected != actual:
                    print('Regular expression is missing a character:')
                    print('Emoji name %r in form %r contains:' % (name, form))
                    print(
                        '\n'.join(
                            [
                                '%r (%r) is not in the regular expression'
                                % (x, x.encode('unicode-escape').decode())
                                for x in pattern.findall(name[1:-1])
                            ]
                        )
                    )

                assert expected == actual, '%s != %s' % (expected, actual)
                assert pattern.search(name[1:-1]) is None


def test_regular_expression_minimal(load_all_languages):  # type:ignore
    # Check that the regular expression emoji.core._EMOJI_NAME_PATTERN only contains the necesseary characters
    from emoji.core import _EMOJI_NAME_PATTERN  # pyright: ignore [reportPrivateUsage]

    pattern_str = '[^%s]' % (_EMOJI_NAME_PATTERN,)
    i = 2
    while i < len(pattern_str) - 1:
        c = pattern_str[i]
        if c == '\\':
            i += 2
            continue
        pattern = re.compile(pattern_str.replace(c, ''))
        failed = False
        for _, emoji_pack in all_language_and_alias_packs():
            for name_in_db in emoji_pack.keys():
                name_in_db = name_in_db[1:-1]
                names = [
                    name_in_db,
                    normalize('NFKC', name_in_db),
                    normalize('NFKD', name_in_db),
                    normalize('NFD', name_in_db),
                    normalize('NFC', name_in_db),
                ]
                for str in names:
                    if pattern.search(str):
                        failed = True
                        break
            if failed:
                break
        if not failed:
            assert failed, 'char: %r is not necessary in regular expression' % (c,)

        i += 1


def test_emojize_complicated_string():
    # A bunch of emoji's with UTF-8 strings to make sure the regex expression is functioning
    name_code = {
        ':flag_for_Ceuta_&_Melilla:': '\U0001f1ea\U0001f1e6',
        ':flag_for_St._Barthélemy:': '\U0001f1e7\U0001f1f1',
        ':flag_for_Côte_d’Ivoire:': '\U0001f1e8\U0001f1ee',
        ':flag_for_Åland_Islands:': '\U0001f1e6\U0001f1fd',
        ':flag_for_São_Tomé_&_Príncipe:': '\U0001f1f8\U0001f1f9',
        ':flag_for_Curaçao:': '\U0001f1e8\U0001f1fc',
    }
    string = ' complicated! '.join(list(name_code.keys()))
    actual = emoji.emojize(string)
    expected = string
    for name, code in name_code.items():
        expected = expected.replace(name, code)
    expected = emoji.emojize(actual)
    assert expected == actual, '%s != %s' % (expected, actual)


def test_emojize_languages(load_all_languages):  # type:ignore
    for lang_code, emoji_pack in all_language_packs():
        for name, emj in emoji_pack.items():
            assert emoji.emojize(name, language=lang_code) == emj


def test_demojize_languages(load_all_languages):  # type:ignore
    for lang_code, emoji_pack in all_language_packs():
        for name, emj in emoji_pack.items():
            assert emoji.demojize(emj, language=lang_code) == name


def test_emojize_variant():
    def remove_variant(s: str) -> str:
        return re.sub('[\ufe0e\ufe0f]$', '', s)

    english_pack = get_emoji_unicode_dict('en')

    assert emoji.emojize(':Taurus:', variant=None) == english_pack[':Taurus:']
    assert emoji.emojize(':Taurus:', variant=None) == emoji.emojize(':Taurus:')
    assert (
        emoji.emojize(':Taurus:', variant='text_type')
        == remove_variant(english_pack[':Taurus:']) + '\ufe0e'
    )
    assert (
        emoji.emojize(':Taurus:', variant='emoji_type')
        == remove_variant(english_pack[':Taurus:']) + '\ufe0f'
    )

    assert (
        emoji.emojize(':admission_tickets:', variant=None)
        == english_pack[':admission_tickets:']
    )
    assert emoji.emojize(':admission_tickets:', variant=None) == emoji.emojize(
        ':admission_tickets:'
    )
    assert (
        emoji.emojize(':admission_tickets:', variant='text_type')
        == remove_variant(english_pack[':admission_tickets:']) + '\ufe0e'
    )
    assert (
        emoji.emojize(':admission_tickets:', variant='emoji_type')
        == remove_variant(english_pack[':admission_tickets:']) + '\ufe0f'
    )

    with suppress_type_checks():  # type:ignore
        with pytest.raises(ValueError):
            emoji.emojize(':admission_tickets:', variant=False)  # type: ignore[arg-type]

        with pytest.raises(ValueError):
            emoji.emojize(':admission_tickets:', variant=True)  # type: ignore[arg-type]

        with pytest.raises(ValueError):
            emoji.emojize(':admission_tickets:', variant='wrong')  # type: ignore[arg-type]

    assert emoji.emojize(':football:') == ':football:'
    assert emoji.emojize(':football:', variant='text_type') == ':football:'
    assert emoji.emojize(':football:', language='alias') == '\U0001f3c8'
    assert (
        emoji.emojize(':football:', variant='emoji_type', language='alias')
        == '\U0001f3c8'
    )


def test_demojize_removes_variant():
    # demojize should remove all variant indicators \ufe0e and \ufe0f from the string
    text = ''.join(
        [
            emoji.emojize(':Taurus:', variant='text_type'),
            emoji.emojize(':Taurus:', variant='emoji_type'),
            emoji.emojize(':admission_tickets:', variant='text_type'),
            emoji.emojize(':admission_tickets:', variant='emoji_type'),
            emoji.emojize(':alien:', variant='text_type'),
            emoji.emojize(':alien:', variant='emoji_type'),
            emoji.emojize(':atom_symbol:', variant='text_type'),
            emoji.emojize(':atom_symbol:', variant='emoji_type'),
        ]
    )

    for lang_code in emoji.LANGUAGES:
        result = emoji.demojize(text, language=lang_code)
        assert '\ufe0e' not in result
        assert '\ufe0f' not in result


def test_emojize_invalid_emoji():
    string = '__---___--Invalid__--__-Name'
    assert emoji.emojize(string) == string

    string = ':: baby:: :_: : : :  : :-: :+:'
    assert emoji.emojize(string) == string


def test_emojize_version(load_all_languages):  # type:ignore
    assert (
        emoji.emojize('Flags like :Belgium: are in version 2.0', version=1.0)
        == 'Flags like  are in version 2.0'
    )
    assert (
        emoji.emojize('Flags like :Belgium: are in version 2.0', version=1.9)
        == 'Flags like  are in version 2.0'
    )
    assert (
        emoji.emojize('Flags like :Belgium: are in version 2.0', version=2.0)
        == 'Flags like 🇧🇪 are in version 2.0'
    )
    assert (
        emoji.emojize('Flags like :Belgium: are in version 2.0', version=3.0)
        == 'Flags like 🇧🇪 are in version 2.0'
    )
    assert (
        emoji.emojize('Boxing gloves :boxing_glove: are in version 3.0', version=0)
        == 'Boxing gloves  are in version 3.0'
    )
    assert (
        emoji.emojize('Boxing gloves :boxing_glove: are in version 3.0', version=1)
        == 'Boxing gloves  are in version 3.0'
    )
    assert (
        emoji.emojize('Boxing gloves :boxing_glove: are in version 3.0', version=2)
        == 'Boxing gloves  are in version 3.0'
    )
    assert (
        emoji.emojize('Boxing gloves :boxing_glove: are in version 3.0', version=3)
        == 'Boxing gloves 🥊 are in version 3.0'
    )
    assert (
        emoji.emojize('Boxing gloves :boxing_glove: are in version 3.0', version=4)
        == 'Boxing gloves 🥊 are in version 3.0'
    )

    assert (
        emoji.emojize(
            'Biking :man_biking: is in 4.0', version=3.0, handle_version='<emoji>'
        )
        == 'Biking <emoji> is in 4.0'
    )
    assert (
        emoji.emojize(
            'Biking :man_biking: is in 4.0',
            version=3.0,
            handle_version=lambda e, data: '<emoji>',
        )
        == 'Biking <emoji> is in 4.0'
    )
    assert (
        emoji.emojize(
            'Biking :man_biking: is in 4.0',
            version=3.0,
            handle_version=lambda e, data: data['fr'],
        )
        == 'Biking :cycliste_homme: is in 4.0'
    )

    def f(emj: str, data: Dict[str, str]) -> str:
        assert data['E'] == 5
        return ''

    assert emoji.emojize(':bowl_with_spoon:', version=-1, handle_version=f) == ''
    assert emoji.emojize(':bowl_with_spoon:') == '\U0001f963'
    assert emoji.emojize(':bowl_with_spoon:', version=4) == ''
    assert emoji.emojize(':bowl_with_spoon:', version=4.9) == ''
    assert emoji.emojize(':bowl_with_spoon:', version=5) == '\U0001f963'
    assert emoji.emojize(':bowl_with_spoon:', version=5.1) == '\U0001f963'
    assert emoji.emojize(':bowl_with_spoon:', version=6) == '\U0001f963'
    assert emoji.emojize(':bowl_with_spoon:', version=4, handle_version='abc') == 'abc'
    assert emoji.emojize(':bowl_with_spoon:', version=4, handle_version=None) == ''
    assert emoji.emojize(':bowl_with_spoon:', version=4, handle_version='') == ''
    assert (
        emoji.emojize(
            ':bowl_with_spoon:', version=4, handle_version=lambda e, d: str(d['E'])
        )
        == '5'
    )


def test_demojize_version():
    assert (
        emoji.emojize('A :T-Rex: is eating a :croissant:', version=3.0)
        == 'A  is eating a 🥐'
    )

    assert (
        emoji.emojize(
            'A :T-Rex: is eating a :croissant:',
            version=3.0,
            handle_version='[Unsupported emoji]',
        )
        == 'A [Unsupported emoji] is eating a 🥐'
    )

    assert (
        emoji.demojize('A 🦖 is eating a 🥐', version=3.0)
        == 'A  is eating a :croissant:'
    )

    assert (
        emoji.demojize('A 🦖 is eating a 🥐', handle_version='X', version=3.0)
        == 'A X is eating a :croissant:'
    )

    assert (
        emoji.demojize('A 🦖 is eating a 🥐', handle_version='X', version=5.0)
        == 'A :T-Rex: is eating a :croissant:'
    )


def test_alias():
    # When lanugage != "alias" aliases should be passed through untouched
    assert emoji.emojize(':soccer:') == ':soccer:'
    assert emoji.emojize(':soccer:', language='alias') == '\U000026bd'
    assert emoji.emojize(':football:') == ':football:'
    assert emoji.emojize(':football:', language='alias') == '\U0001f3c8'
    # Multiple aliases for one emoji:
    assert emoji.emojize(':thumbsup:', language='alias') == emoji.emojize(
        ':+1:', language='alias'
    )
    assert emoji.emojize(':thumbsup:', language='alias') == emoji.emojize(
        ':thumbs_up:', language='alias'
    )
    assert emoji.emojize(':thumbsup:', language='alias') == '\U0001f44d'

    thumbsup = '\U0001f44d'
    assert emoji.demojize(thumbsup, language='alias') != thumbsup
    assert emoji.demojize(thumbsup, language='alias') != ':thumbs_up:'
    assert emoji.demojize(thumbsup, language='alias') != emoji.demojize(thumbsup)

    thailand = '🇹🇭'
    assert emoji.demojize(thailand, language='alias') != thailand
    assert emoji.demojize(thailand, language='alias') != ':Thailand:'
    assert emoji.demojize(thailand, language='alias') != emoji.demojize(thailand)
    assert emoji.demojize(thailand, language='alias', version=1.0) != emoji.demojize(
        thailand, language='alias'
    )

    # No alias
    for emj, emoji_data in emoji.EMOJI_DATA.items():
        if emoji_data['status'] != emoji.STATUS['fully_qualified']:
            continue
        if 'alias' not in emoji_data:
            assert emoji.emojize(emoji_data['en'], language='alias') != emoji_data['en']
            assert emoji.demojize(emj, language='alias') == emoji_data['en']


def test_invalid_alias():
    # Invalid aliases should be passed through untouched
    assert emoji.emojize(':tester:', language='alias') == ':tester:'
    assert emoji.emojize(':footbal:', language='alias') == ':footbal:'
    assert emoji.emojize(':socer:', language='alias') == ':socer:'
    assert emoji.emojize(':socer:', language='alias', variant='text_type') == ':socer:'


def test_demojize_name_only(load_all_languages):  # type:ignore
    for emj, item in emoji.EMOJI_DATA.items():
        if item['status'] != emoji.STATUS['fully_qualified']:
            continue
        for lang_code in emoji.LANGUAGES:
            if lang_code not in item:
                continue
            name = item[lang_code]
            oneway = emoji.emojize(name, language=lang_code)
            assert oneway == emj
            roundtrip = emoji.demojize(oneway, language=lang_code)
            assert name == roundtrip, '%s != %s' % (name, roundtrip)


def test_demojize_complicated_string():
    constructed = 'testing :baby::emoji_modifier_fitzpatrick_type-3: with :eyes: :eyes::eyes: modifiers :baby::emoji_modifier_fitzpatrick_type-5: to symbols ヒㇿ'
    emojid = emoji.emojize(constructed)
    destructed = emoji.demojize(emojid)
    assert constructed == destructed, '%s != %s' % (constructed, destructed)


def test_demojize_delimiters():
    for e in ['\U000026bd', '\U0001f44d', '\U0001f3c8']:
        for d in [
            (':', ':'),
            ('}', '}'),
            ('!$', '!!$'),
            ('[123', '456]'),
            ('😁', '👌'),
            ('[', ']'),
        ]:
            s = emoji.demojize(e, delimiters=d)
            assert s.startswith(d[0])
            assert s.endswith(d[1])

    text = 'Example with an emoji%sin a sentence and %s %s %s %s multiple emoji %s%s%s%s%s in a row'
    for e in ['\U000026bd', '\U0001f44d', '\U0001f3c8']:
        for d in [
            (':', ':'),
            ('{', '}'),
            ('!$', '$!'),
            (':', '::'),
            ('::', '::'),
            ('😁', '👌'),
            ('[', ']'),
        ]:
            print('delimiter: %s' % (d,))
            text_with_unicode = text % ((e,) * 10)
            demojized_text = emoji.demojize(text_with_unicode, delimiters=d)
            assert text_with_unicode != demojized_text, (
                text_with_unicode,
                demojized_text,
            )
            assert e not in demojized_text
            assert emoji.emojize(demojized_text, delimiters=d) == text_with_unicode
            de = emoji.demojize(e, delimiters=d)
            text_with_emoji = text % ((de,) * 10)
            assert demojized_text == text_with_emoji
            assert emoji.emojize(text_with_emoji, delimiters=d) == text_with_unicode


def test_emoji_list():
    assert emoji.emoji_list('Hi, I am 👌 test')[0]['match_start'] == 9
    assert emoji.emoji_list('Hi') == []
    if (
        len('Hello 🇫🇷👌') < 10
    ):  # skip these tests on python with UCS-2 as the string length/positions are different
        assert emoji.emoji_list('Hi, I am fine. 😁') == [
            {'match_start': 15, 'match_end': 16, 'emoji': '😁'}
        ]
        assert emoji.emoji_list('Hello 🇫🇷👌') == [
            {'emoji': '🇫🇷', 'match_start': 6, 'match_end': 8},
            {'emoji': '👌', 'match_start': 8, 'match_end': 9},
        ]


def test_distinct_emoji_list():
    assert emoji.distinct_emoji_list('Hi, I am fine. 😁') == ['😁']
    assert emoji.distinct_emoji_list('Hi') == []
    assert set(emoji.distinct_emoji_list('Hello 🇫🇷👌')) == {'🇫🇷', '👌'}
    assert emoji.distinct_emoji_list('Hi, I am fine. 😁😁😁😁') == ['😁']


def test_emoji_count():
    assert emoji.emoji_count('Hi, I am fine. 😁') == 1
    assert emoji.emoji_count('Hi') == 0
    assert emoji.emoji_count('Hello 🇫🇷👌') == 2
    assert emoji.emoji_count('Hello 🇵🇱🍺🇵🇱', unique=True) == 2


def test_replace_emoji():
    assert emoji.replace_emoji('Hi, I am fine. 😁') == 'Hi, I am fine. '
    assert emoji.replace_emoji('Hi') == 'Hi'
    assert emoji.replace_emoji('Hello 🇫🇷👌') == 'Hello '
    assert emoji.replace_emoji('Hello 🇫🇷👌', 'x') == 'Hello xx'

    def replace(emj: str, data: Dict[str, str]) -> str:
        assert emj in ['🇫🇷', '👌']
        return 'x'

    assert emoji.replace_emoji('Hello 🇫🇷👌', replace) == 'Hello xx'


def test_is_emoji():
    assert emoji.is_emoji('😁')
    assert not emoji.is_emoji('H')
    assert emoji.is_emoji('🇫🇷')
    assert not emoji.is_emoji('🇫🇷🇫🇷')
    assert not emoji.is_emoji('\ufe0f')  # variation selector


def test_long_emoji():
    assert (
        emoji.demojize('This is \U0001f9d1\U0001f3fc\U0000200d\U0001f37c example text')
        == 'This is :person_feeding_baby_medium-light_skin_tone: example text'
    )
    assert (
        emoji.demojize(
            'This is \U0001f468\U0001f3ff\u200d\u2764\ufe0f\u200d\U0001f468\U0001f3ff example text \U0001f469\U0001f3fb\U0000200d\U0001f91d\U0000200d\U0001f468\U0001f3ff'
        )
        == 'This is :couple_with_heart_man_man_dark_skin_tone: example text :woman_and_man_holding_hands_light_skin_tone_dark_skin_tone:'
    )
    assert (
        emoji.demojize(
            'This is \U0001f468\U0001f3ff\u200d\u2764\ufe0f\u200d\U0001f468\U0001f3ff\U0001f468\U0001f3ff\u200d\u2764\ufe0f\u200d\U0001f48b\u200d\U0001f468\U0001f3ff example text \U0001f469\U0001f3fb\U0000200d\U0001f91d\U0000200d\U0001f468\U0001f3ff'
        )
        == 'This is :couple_with_heart_man_man_dark_skin_tone::kiss_man_man_dark_skin_tone: example text :woman_and_man_holding_hands_light_skin_tone_dark_skin_tone:'
    )
    assert (
        emoji.demojize(
            '\U0001f46b\U0001f3fb This is \U0001f468\U0001f3ff\U0001f468\U0001f3ff\u200d\u2764\ufe0f\u200d\U0001f468\U0001f3ff\U0001f468\U0001f3ff\u200d\u2764\ufe0f\u200d\U0001f48b\u200d\U0001f468\U0001f3ff example text \U0001f469\U0001f3fb\U0000200d\U0001f91d\U0000200d\U0001f468\U0001f3ff'
        )
        == ':woman_and_man_holding_hands_light_skin_tone: This is :man_dark_skin_tone::couple_with_heart_man_man_dark_skin_tone::kiss_man_man_dark_skin_tone: example text :woman_and_man_holding_hands_light_skin_tone_dark_skin_tone:'
    )
    assert (
        emoji.demojize(
            '\U0001f46b\U0001f3fb\U0001f46b\U0001f3fb\U0001f469\U0001f3fb\U0000200d\U0001f91d\U0000200d\U0001f468\U0001f3ff\U0001faf1\U0001f3fd\U0001faf1\U0001f3fd\U0000200d\U0001faf2\U0001f3ff'
        )
        == ':woman_and_man_holding_hands_light_skin_tone::woman_and_man_holding_hands_light_skin_tone::woman_and_man_holding_hands_light_skin_tone_dark_skin_tone::rightwards_hand_medium_skin_tone::handshake_medium_skin_tone_dark_skin_tone:'
    )
    s = ':crossed_fingers_medium-light_skin_tone::crossed_fingers::crossed_fingers_dark_skin_tone:'
    assert emoji.demojize(emoji.demojize(s)) == s


def test_untranslated(load_all_languages):  # type:ignore
    for item in emoji.EMOJI_DATA.values():
        if item['status'] != emoji.STATUS['fully_qualified']:
            continue
        if 'es' not in item:
            # untranslated
            value = emoji.emojize(item['en'], language='en')
            roundtrip = emoji.demojize(value, language='es')
            assert roundtrip == value, '%s != %s (from %s)' % (
                ascii(roundtrip),
                ascii(value),
                item['en'],
            )
        else:
            # translated
            value = emoji.emojize(item['en'], language='en')
            roundtrip = emoji.demojize(value, language='es')
            assert roundtrip == item['es'], '%s != %s' % (roundtrip, item['es'])


def test_text(load_all_languages):  # type:ignore
    emoji.config.demojize_keep_zwj = True  # Restore default config value
    emoji.config.replace_emoji_keep_zwj = False  # Restore default config value

    UCS2 = len('Hello 🇫🇷👌') > 9  # don't break up characters on python with UCS-2

    text = """Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur.
Excepteur sint occaecat in reprehenderit in cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Stróż pchnął kość w quiz gędźb vel fax myjń.
Høj bly gom vandt fræk sexquiz på wc.
Съешь же ещё этих мягких французских булок, да выпей чаю.
За миг бях в чужд плюшен скърцащ фотьойл.
هلا سكنت بذي ضغثٍ فقد زعموا — شخصت تطلب ظبياً راح مجتازا
שפן אכל קצת גזר בטעם חסה, ודי
ऋषियों को सताने वाले दुष्ट राक्षसों के राजा रावण का सर्वनाश करने वाले विष्णुवतार भगवान श्रीराम, अयोध्या के महाराज दशरथ के बड़े सपुत्र थे।
とりなくこゑす ゆめさませ みよあけわたる ひんかしを そらいろはえて おきつへに ほふねむれゐぬ もやのうち
視野無限廣，窗外有藍天
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur.
Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
"""

    def default_select(emj_data: Dict[str, Any]) -> str:
        return emj_data['en']

    def add_random_emoji(
        text: str,
        lst: List[Tuple[str, Dict[str, Any]]],
        select: Callable[[Dict[str, Any]], Union[str, Literal[False]]] = default_select,
    ) -> Tuple[str, str, List[str]]:
        emoji_list: List[str] = []
        text_with_unicode = ''
        text_with_placeholder = ''
        for i in range(0, len(text), 10):
            while True:
                emj, emj_data = random.choice(lst)
                placeholder = select(emj_data)
                if placeholder:
                    break

            if UCS2:
                j = text.find(' ', i, i + 10)
                if j == -1:
                    continue
            else:
                j = random.randint(i, i + 10)

            text_with_unicode += text[i:j]
            text_with_unicode += emj
            text_with_unicode += text[j : i + 10]

            text_with_placeholder += text[i:j]
            text_with_placeholder += placeholder
            text_with_placeholder += text[j : i + 10]

            emoji_list.append(emj)

        return text_with_unicode, text_with_placeholder, emoji_list

    def clean(s: str) -> str:
        return s.replace('\u200d', '').replace('\ufe0f', '')

    qualified_emoji_list = [
        (emj, item)
        for emj, item in emoji.EMOJI_DATA.items()
        if item['status'] == emoji.STATUS['fully_qualified']
    ]

    all_emoji_list_except_component = [
        (emj, item)
        for emj, item in emoji.EMOJI_DATA.items()
        if item['status'] >= emoji.STATUS['fully_qualified']
    ]

    # qualified emoji
    text_with_unicode, text_with_placeholder, emoji_list = add_random_emoji(
        text, qualified_emoji_list
    )
    assert emoji.demojize(text_with_unicode) == text_with_placeholder
    assert emoji.emojize(text_with_placeholder) == text_with_unicode
    if not UCS2:
        assert emoji.replace_emoji(text_with_unicode, '') == text
    assert set(emoji.distinct_emoji_list(text_with_unicode)) == set(emoji_list)
    for i, lis in enumerate(emoji.emoji_list(text_with_unicode)):
        assert lis['emoji'] == emoji_list[i]

    # qualified emoji from "es"
    def select_es(emj_data: Dict[str, Any]) -> Union[str, Literal[False]]:
        return emj_data['es'] if 'es' in emj_data else False

    text_with_unicode, text_with_placeholder, emoji_list = add_random_emoji(
        text, qualified_emoji_list, select=select_es
    )
    assert emoji.demojize(text_with_unicode, language='es') == text_with_placeholder
    assert emoji.emojize(text_with_placeholder, language='es') == text_with_unicode
    if not UCS2:
        assert emoji.replace_emoji(text_with_unicode, '') == text
    assert set(emoji.distinct_emoji_list(text_with_unicode)) == set(emoji_list)
    for i, lis in enumerate(emoji.emoji_list(text_with_unicode)):
        assert lis['emoji'] == emoji_list[i]

    # qualified emoji from "alias"
    def select_alias(emj_data: Dict[str, Any]) -> Union[str, Literal[False]]:
        return emj_data['alias'][0] if 'alias' in emj_data else False

    text_with_unicode, text_with_placeholder, emoji_list = add_random_emoji(
        text, qualified_emoji_list, select=select_alias
    )
    assert emoji.demojize(text_with_unicode, language='alias') == text_with_placeholder
    assert emoji.emojize(text_with_placeholder, language='alias') == text_with_unicode
    if not UCS2:
        assert emoji.replace_emoji(text_with_unicode, '') == text
    assert set(emoji.distinct_emoji_list(text_with_unicode)) == set(emoji_list)
    for i, lis in enumerate(emoji.emoji_list(text_with_unicode)):
        assert lis['emoji'] == emoji_list[i]

    # all emoji (except components)
    text_with_unicode, text_with_placeholder, emoji_list = add_random_emoji(
        text, all_emoji_list_except_component
    )
    assert emoji.demojize(text_with_unicode) == text_with_placeholder
    assert clean(emoji.emojize(text_with_placeholder)) == clean(text_with_unicode)
    if not UCS2:
        assert emoji.replace_emoji(text_with_unicode, '') == text
    assert set(emoji.distinct_emoji_list(text_with_unicode)) == set(emoji_list)
    for i, lis in enumerate(emoji.emoji_list(text_with_unicode)):
        assert lis['emoji'] == emoji_list[i]


def test_text_multiple_times(load_all_languages: Any):
    # Run test_text() multiple times because it relies on a random text
    for _ in range(100):
        test_text(load_all_languages)


def test_invalid_chars():
    invalidchar = '\U0001f20f'
    assert emoji.demojize(invalidchar) == invalidchar, '%r != %r' % (
        ascii(emoji.demojize(invalidchar)),
        ascii(invalidchar),
    )
    assert emoji.demojize(invalidchar) == invalidchar, '%r != %r' % (
        ascii(emoji.demojize(invalidchar)),
        ascii(invalidchar),
    )

    invalidchar = 'u\2302 ⌂'
    assert emoji.demojize(invalidchar) == invalidchar, '%r != %r' % (
        ascii(emoji.demojize(invalidchar)),
        ascii(invalidchar),
    )
    assert emoji.demojize(invalidchar) == invalidchar, '%r != %r' % (
        ascii(emoji.demojize(invalidchar)),
        ascii(invalidchar),
    )


def test_combine_with_component():
    text = 'Example of a combined emoji%sin a sentence'

    combined = emoji.emojize(text % ':woman_dark_skin_tone:')
    separated = emoji.emojize(text % ':woman::dark_skin_tone:')
    assert combined == separated, '%r != %r' % (ascii(combined), ascii(separated))

    combined = emoji.emojize(text % ':woman_dark_skin_tone_white_hair:')
    separated = emoji.emojize(text % ':woman::dark_skin_tone:\u200d:white_hair:')
    assert combined == separated, '%r != %r' % (ascii(combined), ascii(separated))


purely_emoji_testdata = [
    ('\U0001f600\ufe0f', True),
    ('\U0001f600', True),
    ('\U0001f600\U0001f600\U0001f600', True),
    ('abc', False),
    ('abc\U0001f600', False),
    ('\U0001f600c', False),
    ('\u270a\U0001f3fe', True),
]


@pytest.mark.parametrize('string,expected', purely_emoji_testdata)
def test_purely_emoji(string: str, expected: bool):
    assert emoji.purely_emoji(string) == expected



import sys
import functools

import pytest

import emoji


@pytest.fixture
def clean_module():
    """Ensures a fresh import and init of the emoji module.
    Unloads all emoji.* modules, then loads the emoji module,
    runs the test function, then deletes all the emoji.* modules
    """
    global emoji
    for name in [
        name
        for name in sys.modules
        if name.startswith('emoji') or name.startswith('test')
    ]:
        del sys.modules[name]
    import emoji

    # Increase cache size to avoid cache misses during tests
    emoji.unicode_codes.get_emoji_by_name = functools.lru_cache(maxsize=None)(  # type:ignore
        emoji.unicode_codes.get_emoji_by_name.__wrapped__  # type:ignore
    )

    yield
    for name in [
        name
        for name in sys.modules
        if name.startswith('emoji') or name.startswith('test')
    ]:
        del sys.modules[name]

    # Increase cache size to avoid cache misses during tests
    emoji.unicode_codes.get_emoji_by_name = functools.lru_cache(maxsize=None)(  # type:ignore
        emoji.unicode_codes.get_emoji_by_name.__wrapped__  # type:ignore
    )


def test_language_loaded_after_emojize(clean_module):  # type:ignore
    emoji.emojize('string', language='es')
    assert 'es' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]


def test_language_loaded_after_demojize(clean_module):  # type:ignore
    emoji.demojize('string', language='es')
    assert 'es' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]


def test_language_not_auto_loaded(clean_module):  # type:ignore
    assert 'es' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]


def test_key_access_load_language(clean_module):  # type:ignore
    assert 'es' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    with pytest.deprecated_call():
        assert emoji.EMOJI_DATA[emoji.emojize(':lion:')]['es']
    assert 'es' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    assert 'es' in emoji.EMOJI_DATA[emoji.emojize(':zebra:')]

    with pytest.raises(KeyError):
        emoji.EMOJI_DATA[emoji.emojize(':lion:')]['xyz']


def test_explicit_load_language(clean_module):  # type:ignore
    assert 'fr' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    emoji.config.load_language('fr')
    assert 'fr' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    assert emoji.EMOJI_DATA[emoji.emojize(':lion:')]['fr'] == ':tête_de_lion:'

    assert 'es' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    emoji.config.load_language('es')
    assert 'es' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]


def test_explict_load_all_languages(clean_module):  # type:ignore
    assert 'fr' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    assert 'es' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    emoji.config.load_language()
    assert 'es' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    assert 'fr' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    assert emoji.EMOJI_DATA[emoji.emojize(':lion:')]['fr'] == ':tête_de_lion:'


def test_load_language_multiple_times(clean_module):  # type:ignore
    assert 'fr' not in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    for _ in range(10000):
        emoji.config.load_language('fr')
    assert 'fr' in emoji.EMOJI_DATA[emoji.emojize(':lion:')]
    assert emoji.EMOJI_DATA[emoji.emojize(':lion:')]['fr'] == ':tête_de_lion:'


unsupported_languages = ['', ' ', '1', 'e', 'z', 'zz', 'ZZ', 'EN', 'ES', 'xyz', 'eses']


@pytest.mark.parametrize('lang', unsupported_languages)
def test_load_unsupported_language(lang: str):
    with pytest.raises(NotImplementedError):
        emoji.config.load_language(lang)


@pytest.mark.parametrize('lang', unsupported_languages)
def test_access_unsupported_language(lang: str):
    with pytest.raises(KeyError):
        emoji.EMOJI_DATA[emoji.emojize(':lion:')][lang]


@pytest.mark.parametrize('lang', unsupported_languages)
def test_emojize_unsupported_language(lang: str):
    with pytest.raises(NotImplementedError):
        emoji.emojize(':lion:', language=lang)

    with pytest.raises(NotImplementedError):
        emoji.demojize(':lion:', language=lang)
